
# Python3.10 moved this module, support both import paths
try:
    from collections import Iterable
except ImportError:
    from collections.abc import Iterable

import itertools
import re
import pathlib

import jinja2
import markupsafe
from flask import abort, redirect, render_template, request
from flask.helpers import url_for
from flask.views import View
from werkzeug.urls import Href

from puncover import collector
from puncover.backtrace_helper import BacktraceHelper

KEY_OUTPUT_FILE_NAME = "output_file_name"

def renderer_from_context(context):
    if isinstance(context, HTMLRenderer):
        return context
    if context:
        return context.parent.get("renderer", None)

    return None

def symbol_file(value):
    return value.get(collector.BASE_FILE, '__builtin')

@jinja2.pass_context
def symbol_url_filter(context, value):
    renderer = renderer_from_context(context)
    if renderer:
        return renderer.url_for_symbol(value)

    return None

@jinja2.pass_context
def symbol_file_url_filter(context, value):
    f = value.get(collector.FILE, None)
    return symbol_url_filter(context, f) if f else None

def none_sum(a, b):
    if a is not None:
        return a + b if b is not None else a
    return b

def symbol_traverse(s, func):
    if isinstance(s, list):
        result = None
        for si in [symbol_traverse(i, func) for i in s]:
            if si is not None:
                result = none_sum(result, si)
        return result

    if collector.TYPE in s:
        if s[collector.TYPE] == collector.TYPE_FILE:
            return sum([symbol_traverse(s, func) for s in s[collector.SYMBOLS]])
        if s[collector.TYPE] == collector.FOLDER:
            return sum([symbol_traverse(s, func) for s in itertools.chain(s[collector.SUB_FOLDERS], s[collector.FILES])])

    return func(s)

def traverse_filter_wrapper(value, func):
    result = symbol_traverse(value, func)
    return result if result != 0 else ""

@jinja2.pass_context
def symbol_code_size_filter(context, value):
    return traverse_filter_wrapper(value, lambda s: s.get(collector.SIZE, None) if s.get(collector.TYPE, None) == collector.TYPE_FUNCTION else 0)

@jinja2.pass_context
def symbol_var_size_filter(context, value):
    return traverse_filter_wrapper(value, lambda s: s.get(collector.SIZE, None) if s.get(collector.TYPE, None) == collector.TYPE_VARIABLE else 0)

@jinja2.pass_context
def symbol_stack_size_filter(context, value, stack_base=None):
    if isinstance(stack_base, str):
        stack_base = None
    result = traverse_filter_wrapper(value, lambda s: s.get(collector.STACK_SIZE, None) if s.get(collector.TYPE, None) == collector.TYPE_FUNCTION else None)
    return none_sum(result, stack_base)

@jinja2.pass_context
def if_not_none_filter(context, value, default_value=""):
    return value if value is not None else default_value

@jinja2.pass_context
def unique_filter(context, value):
    if isinstance(value, Iterable):
        result = []
        for v in value:
            if v not in result:
                result.append(v)
        return result

    return value


@jinja2.pass_context
def assembly_filter(context, value):
    renderer = context.parent.get("renderer", None)
    def linked_symbol_name(name):
        if context:
            display_name = renderer.display_name_for_symbol_name(name)
            url = renderer.url_for_symbol_name(name, context)
            if url:
                return '<a href="%s">%s</a>' % (url, display_name)
        return name

    value = str(value)

    # Get a clean display name - and a URL - for symbol names in comments
    #   b8:	f000 f8de 	bleq	278 &lt:__aeabi_dmul+0x1dc&gt:
    pattern = re.compile(r"&lt;(\w+)")
    s = pattern.sub(lambda match: "&lt;"+linked_symbol_name(match.group(1)), value)

    # Get a clean display name for symbol names in labels
    # _ZN6Stream9readBytesEPcj():
    # FIXME: Unfortunately symbols that have been inlined will not be in our global
    # symbol name list and we will not be able to unmangle them (this is only a problem
    # for c++ symbols).
    pattern = re.compile(r"^(_.*)\(\):$")

    def display_name_for_label(match):
        display_name = renderer.display_name_for_symbol_name(match.group(1))

        if display_name.endswith(")"):
            # C++ symbols will include parenthesis and arguments
            return display_name + ":"
        else:
            # Other symbols will just have a name
            return display_name + "():"
    s = pattern.sub(display_name_for_label, s)

    return s
    # return str("&lt;")

@jinja2.pass_context
def symbols_filter(context, value):
    renderer = renderer_from_context(context)

    if renderer:
        helper = BacktraceHelper(renderer.collector)

        def make_links(s):
            name = s[collector.NAME]
            url = renderer.url_for_symbol_name(name, context)
            return '<a href="%s">%s</a>' % (url, name)

        if isinstance(value, markupsafe.Markup):
            value = value.__html__().unescape()
        return helper.transform_known_symbols(value, make_links)

    return value

@jinja2.pass_context
def chain_filter(context, value, second_value=None):
    return list(itertools.chain(value, second_value if second_value else []))


def is_int_ge(x, cmp):
    return isinstance(x, int) and x >= cmp


@jinja2.pass_context
def bytes_filter(context, x):
    if not is_int_ge(x, 0):
        return x

    result = ''
    while x >= 1000:
        x, r = divmod(x, 1000)
        result = '<span class="secondary">,</span>%03d%s' % (r, result)
    return "%d%s" % (x, result)


@jinja2.pass_context
def style_background_bar_filter(context, x, total, color=None):
    if not is_int_ge(x, 1) or not is_int_ge(total, 1):
        return ''

    if color is None:
        color = 'rgba(0,0,255,0.07)'

    x = min(x, total)
    percent = 100 * x // total
    return 'background:linear-gradient(90deg, {1} {0}%, transparent {0}%);'.format(percent, color)

@jinja2.pass_context
def col_sortable_filter(context, title, is_alpha=False, id=None):

    id = title if id is None else id
    id = id.lower()

    # when sorting for numbers, we're interested in large numbers first
    next_sort = 'asc' if is_alpha else 'desc'
    sort_id, sort_order = context.parent.get('sort', 'a_b').split('_')
    classes = ['sortable']
    if sort_id.lower() == id:
        sort_class = 'sort_' + sort_order
        next_sort = 'desc' if sort_order == 'asc' else 'asc'
        if is_alpha:
            sort_class += '_alpha'
        classes.append(sort_class)

    next_sort = id + '_' + next_sort

    # replace/set ?sort= in URL
    args = request.args.copy()
    args['sort'] = next_sort
    url = Href(request.base_url, sort=True)

    return '<a href="%s" class="%s">%s</a>' % (url(args), ' '.join(classes), title)


@jinja2.pass_context
def sorted_filter(context, symbols):
    sort_id, sort_order = context.parent['sort'].split('_')

    def to_num(v):
        if v is None or v == '':
            return 0
        return int(v)

    key = {
        'name': lambda e: e.get(collector.DISPLAY_NAME, e.get(collector.NAME, None)).lower(),
        'code': lambda e: to_num(symbol_code_size_filter(context, e)),
        'stack': lambda e: to_num(symbol_stack_size_filter(context, e)),
        'vars': lambda e: to_num(symbol_var_size_filter(context, e)),
    }[sort_id]

    return list(sorted(symbols, key=key, reverse=(sort_order == 'desc')))


class HTMLRenderer(View):

    def __init__(self, collector):
        self.collector = collector
        self.template_vars = {
            "renderer": self,
            "SLASH": '<span class="slash">/</span>',
            "root_folders": list(collector.root_folders()),
            "sort": 'name_asc',
            "all_symbols": collector.all_symbols(),
            "all_functions": collector.all_functions(),
            "all_variables": collector.all_variables(),
        }

    def render_template(self, template_name, file_name):
        self.template_vars['sort'] = request.args.get('sort', 'name_asc')
        self.template_vars['request'] = request
        self.template_vars[KEY_OUTPUT_FILE_NAME] = file_name
        return render_template(template_name, **self.template_vars)

    def url_for_symbol_name(self, name, context=None):
        symbol = self.collector.symbol(name, False)
        return symbol_url_filter(context, symbol) if symbol else None

    def display_name_for_symbol_name(self, name):
        symbol = self.collector.symbol(name, False)
        return symbol['display_name'] if symbol else name

    def url_for(self, endpoint, **values):
        result = url_for(endpoint, **values)
        href = Href(result)
        # pass along any query parameters
        # this is kind of hacky as it replaces any existing parameters
        return href(request.args)

    def url_for_symbol(self, value):
        if value[collector.TYPE] in [collector.TYPE_FUNCTION]:
            return self.url_for("path", path=self.collector.qualified_symbol_name(value))

        # file or folder
        path = value.get(collector.PATH, None)
        return self.url_for("path", path=path) if path else ""


    def dispatch_request(self):
        return self.render_template(self.template, "index.html")


class OverviewRenderer(HTMLRenderer):

    def dispatch_request(self):
        return self.render_template("overview.html.jinja", "index.html")


class PathRenderer(HTMLRenderer):

    def dispatch_request(self, path=None):
        if path.endswith("/"):
            path = path[:-1]

        generic_path = pathlib.Path(path)
        symbol = self.collector.symbol(path)

        if symbol:
            self.template_vars["symbol"] = symbol
            return self.render_template("symbol.html.jinja", "symbol")

        file_element = self.collector.file_elements.get(generic_path, None)
        if file_element and file_element[collector.TYPE] == collector.TYPE_FILE:
            self.template_vars["file"] = file_element
            return self.render_template("file.html.jinja", path)
        elif file_element and file_element[collector.TYPE] == collector.TYPE_FOLDER:
            self.template_vars["folder"] = file_element
            return self.render_template("folder.html.jinja", path)

        print("### " + path)
        for f in sorted([f[collector.PATH] for f in self.collector.file_elements.values()]):
            print(f)
        # print "## root folders"
        # for f in self.collector.root_folders():
        #     print f[collector.PATH]
        # print "## collapsed root folders"
        # for f in self.collector.collapsed_root_folders():
        #     print f[collector.PATH]

        abort(404)


class SymbolRenderer(HTMLRenderer):

    def dispatch_request(self, symbol_name=None):
        symbol = self.collector.symbol(symbol_name, qualified=False)
        if not symbol:
            abort(404)

        return redirect(self.url_for("path", path=self.collector.qualified_symbol_name(symbol)))

class AllSymbolsRenderer(HTMLRenderer):

    def dispatch_request(self, symbol_name=None):
        return self.render_template("all_symbols.html.jinja", "all")


class RackRenderer(HTMLRenderer):

    def dispatch_request(self, symbol_name=None):
        if request.method == "POST":
            helper = BacktraceHelper(self.collector)

            snippet = request.form["snippet"]
            self.template_vars["snippet"] = snippet
            self.template_vars["functions"] = helper.derive_function_symbols(snippet)

        return self.render_template("rack.html.jinja", "rack")


def register_jinja_filters(jinja_env):
    jinja_env.filters["symbol_url"] = symbol_url_filter
    jinja_env.filters["symbol_file_url"] = symbol_file_url_filter
    jinja_env.filters["symbol_code_size"] = symbol_code_size_filter
    jinja_env.filters["symbol_var_size"] = symbol_var_size_filter
    jinja_env.filters["symbol_stack_size"] = symbol_stack_size_filter
    jinja_env.filters["if_not_none"] = if_not_none_filter
    jinja_env.filters["unique"] = unique_filter
    jinja_env.filters["assembly"] = assembly_filter
    jinja_env.filters["symbols"] = symbols_filter
    jinja_env.filters["chain"] = chain_filter
    jinja_env.filters["bytes"] = bytes_filter
    jinja_env.filters["style_background_bar"] = style_background_bar_filter
    jinja_env.filters["col_sortable"] = col_sortable_filter
    jinja_env.filters["sorted"] = sorted_filter



def register_urls(app, collector):
    app.add_url_rule("/", view_func=OverviewRenderer.as_view("overview", collector=collector))
    app.add_url_rule("/all/", view_func=AllSymbolsRenderer.as_view("all", collector=collector))
    app.add_url_rule("/path/<path:path>/", view_func=PathRenderer.as_view("path", collector=collector))
    app.add_url_rule("/symbol/<string:symbol_name>", view_func=SymbolRenderer.as_view("symbol", collector=collector))
    app.add_url_rule("/rack/", view_func=RackRenderer.as_view("rack", collector=collector), methods=["GET", "POST"])




class BuilderMiddleware(object):
    def __init__(self, app, builder):
        self.app = app
        self.builder = builder

    def __call__(self, environ, start_response):
        self.builder.build_if_needed()
        return self.app(environ, start_response)


import fnmatch
import os
import re
import sys
import pathlib

NAME = "name"
DISPLAY_NAME = "display_name"
SIZE = "size"
PATH = "path"
BASE_FILE = "base_file"
LINE = "line"
ASM = "asm"
STACK_SIZE = "stack_size"
STACK_QUALIFIERS = "stack_qualifiers"
ADDRESS = "address"
TYPE = "type"
TYPE_FUNCTION = "function"
TYPE_VARIABLE = "variable"
TYPE_FILE = "file"
TYPE_FOLDER = "folder"
PREV_FUNCTION = "prev_function"
NEXT_FUNCTION = "next_function"
FUNCTIONS = "functions"
VARIABLES = "variables"
SYMBOLS = "symbols"
FILE = "file"
FILES = "files"
FOLDER = "folder"
ROOT = "root"
ANCESTORS = "ancestors"
SUB_FOLDERS = "sub_folders"
COLLAPSED_NAME = "collapsed_name"
COLLAPSED_SUB_FOLDERS = "collapsed_sub_folders"
CALLEES = "callees"
CALLERS = "callers"

DEEPEST_CALLEE_TREE = "deepest_callee_tree"
DEEPEST_CALLER_TREE = "deepest_caller_tree"

PYTHON_VER = {"major": sys.version_info[0], "minor": sys.version_info[1]}

def warning(*objs):
    print("WARNING: ", *objs, file=sys.stderr)


def left_strip_from_list(lines):
    if len(lines) <= 0:
        return lines

    # detect longest common sequence of white spaces
    longest_match = re.match(r"^\s*", lines[0]).group(0)


    for line in lines:
        while not line.startswith(longest_match):
            longest_match = longest_match[:-1]

    # remove from each string
    return list([line[len(longest_match):] for line in lines])


class Collector:

    def __init__(self, gcc_tools):
        self.gcc_tools = gcc_tools
        self.symbols = {}
        self.file_elements = {}
        self.symbols_by_qualified_name = None
        self.symbols_by_name = None

    def reset(self):
        self.symbols = {}
        self.file_elements = {}
        self.symbols_by_qualified_name = None
        self.symbols_by_name = None

    def qualified_symbol_name(self, symbol):
        if BASE_FILE in symbol:
            html_path = pathlib.Path.joinpath(symbol[PATH], symbol[NAME])
            return str(html_path)
        return symbol[NAME]

    def symbol(self, name, qualified=True):
        self.build_symbol_name_index()
        index = self.symbols_by_qualified_name if qualified else self.symbols_by_name
        return index.get(name, None)

    def symbol_by_addr(self, addr):
        int_addr = int(addr, 16)
        return self.symbols.get(int_addr, None)

    def add_symbol(self, name, address, size=None, file=None, line=None, assembly_lines=None, type=None, stack_size=None):
        int_address = int(address, 16)
        sym = self.symbols.get(int_address, {})
        if NAME in sym and sym[NAME] != name:
            # warning("Name for symbol at %s inconsistent (was '%s', now '%s')" % (address, sym[NAME], name))
            pass
        else:
            sym[NAME] = name
        if size:
            sym[SIZE] = int(size)
        if file:
            sym[PATH] = pathlib.Path(file)
            sym[BASE_FILE] = sym[PATH].name
        if line:
            sym[LINE] = line
        if assembly_lines:
            assembly_lines = left_strip_from_list(assembly_lines)

            sym[ASM] = assembly_lines
            sym[TYPE] = TYPE_FUNCTION
        if type:
            sym[TYPE] = type
        if stack_size:
            sym[STACK_SIZE] = stack_size

        sym[ADDRESS] = address

        self.symbols[int_address] = sym
        return sym

    # 00000550 00000034 T main	/Users/behrens/Documents/projects/pebble/puncover/puncover/build/../src/puncover.c:25
    if os.name == 'nt':
        parse_size_line_re = re.compile(r"^([\da-f]{8})\s+([\da-f]{8})\s+(.)\s+(\w+)(\s+([a-zA-Z]:.+)):(\d+)?")
    else:
        parse_size_line_re = re.compile(r"^([\da-f]{8})\s+([\da-f]{8})\s+(.)\s+(\w+)(\s+([^:]+):(\d+))?")


    def parse_size_line(self, line):
        # print(line)
        match = self.parse_size_line_re.match(line)
        if not match:
            return False

        addr = match.group(1)
        size = int(match.group(2), 16)
        type = match.group(3)
        name = match.group(4)
        if match.group(5):
            file = match.group(6)
            line = int(match.group(7))
        else:
            file = None
            line = None

        types = {"T": TYPE_FUNCTION, "D": TYPE_VARIABLE, "B": TYPE_VARIABLE, "R": TYPE_VARIABLE}

        self.add_symbol(name, address=addr, size=size, file=file, line=line, type = types.get(type.upper(), None))

        return True

    # 00000098 <pbl_table_addr>:
    # 00000098 <pbl_table_addr.constprop.0>:
    parse_assembly_text_function_start_pattern = re.compile(r"^([\da-f]{8})\s+<(\.?\w+)(\..*)?>:")

    # /Users/behrens/Documents/projects/pebble/puncover/pebble/build/../src/puncover.c:8
    parse_assembly_text_c_reference_pattern = re.compile(r"^(/[^:]+)(:(\d+))?")

    def parse_assembly_text(self, assembly):
        # print(assembly)
        name = None
        addr = None
        symbol_file = None
        symbol_line = None
        assembly_lines = []
        found_symbols = 0

        def flush_current_symbol():
            if name and addr:
                self.add_symbol(name, addr, assembly_lines=assembly_lines, file=symbol_file, line=symbol_line)
                return 1
            return 0

        for line in assembly.split("\n"):
            match = self.parse_assembly_text_function_start_pattern.match(line)
            if match:
                found_symbols += flush_current_symbol()
                addr = match.group(1)
                name = match.group(2)
                symbol_file = None
                symbol_line = None
                assembly_lines = []
            else:
                file_match = self.parse_assembly_text_c_reference_pattern.match(line)
                if not file_match and line.strip() != "":
                    assembly_lines.append(line)
                elif file_match and not symbol_file:
                    symbol_file = file_match.group(1)
                    if file_match.group(3):
                        symbol_line = int(file_match.group(3))

        found_symbols += flush_current_symbol()
        return found_symbols

    # puncover.c:8:43:dynamic_stack2	16	dynamic
    # puncover.c:14:40:0	16	dynamic,bounded
    # puncover.c:8:43:dynamic_stack2	16	dynamic
    parse_stack_usage_line_pattern = re.compile(r"^(.*?\.[ch](pp)?):(\d+):(\d+):([^\t]+)\t+(\d+)\s+([a-z,]+)")

    def parse_stack_usage_line(self, line):
        match = self.parse_stack_usage_line_pattern.match(line)
        if not match:
            return False

        file = pathlib.Path(match.group(1))
        base_file_name = file.name
        line = int(match.group(3))
        symbol_name = match.group(5)
        stack_size = int(match.group(6))
        stack_qualifier = match.group(7)

        return self.add_stack_usage(base_file_name, line, symbol_name, stack_size, stack_qualifier)

    # TODO: handle operators, e.g. String::operator=(char const*)
    # TODO: handle templates, e.g. void LinkedList<T>::clear() [with T = Loggable]
    re_cpp_display_name = re.compile(r"^(\w[^\(\s]*\s)*(\w+::~?)?(\w+)(\([^\)]*\))?(\sconst)?$")

    def display_name_simplified(self, name):
        # .su files have elements such as "virtual size_t Print::write(const uint8_t*, size_t)"
        # c++filt gives us "Print::write(unsigned char const*, unsigned int)"

        m = self.re_cpp_display_name.match(name)
        if m:
            groups = list(m.groups(''))

            def replace_identifiers(m):
                # these values were derived from an ARM 32Bit target
                # it could be that they need further adjustments
                # yes, we are treating int as long works only for 32bit platforms
                # right now, our sample projects use both types unpredictably in the same binary (oh, dear)
                mapping = {
                    'const': '', # we ignore those as a feasible simplification
                    'size_t': 'unsigned long',
                    'uint8_t': 'unsigned char',
                    'int8_t': 'signed char',
                    'uint16_t': 'unsigned short',
                    'int16_t': 'short',
                    'uint32_t': 'unsigned long',
                    'int32_t': 'long',
                    'uint64_t': 'unsigned long long',
                    'int64_t': 'long long',
                    'byte': 'unsigned char',
                    'int': 'long',
                }

                return mapping.get(m.group(), m.group())

            # in case, we have parameters, simplify those
            groups[3] = re.sub(r'\w+', replace_identifiers, groups[3])

            # TODO: C allows you to write the same C types in many different notations
            # http://ieng9.ucsd.edu/~cs30x/Std.C/types.html#Basic%20Integer%20Types
            # applies to tNMEA2000::SetProductInformation or Print::printNumber

            # remove leading "virtual size_t" etc.
            # non-matching groups should be empty strings
            name = ''.join(groups[1:])

        # remove white space artifacts from previous replacements
        for k, v in [('   ', ' '), ('  ', ' '), ('( ', '('), (' )', ')'), ('< ', '<'), (' >', '>'), (' *', '*'), (' &', '&')]:
            name = name.replace(k, v)

        return name

    def display_names_match(self, a, b):
        if a is None or b is None:
            return False

        if a == b:
            return True

        simplified_a = self.display_name_simplified(a)
        simplified_b = self.display_name_simplified(b)
        return simplified_a == simplified_b

    def add_stack_usage(self, base_file_name, line, symbol_name, stack_size, stack_qualifier):
        basename_symbols = [s for s in self.symbols.values() if s.get(BASE_FILE, None) == base_file_name]
        for symbol in basename_symbols:

            if symbol.get(LINE, None) == line or self.display_names_match(symbol_name, symbol.get(DISPLAY_NAME, None)):
                symbol[STACK_SIZE] = stack_size
                symbol[STACK_QUALIFIERS] = stack_qualifier
                return True

        # warning("Couldn't find symbol for %s:%d:%s" % (base_file_name, line, symbol_name))
        return False

    windows_path_pattern = re.compile(r"^([a-zA-Z]+)(:)(\\)(.+)$")
    
    def normalize_files_paths(self, base_dir):
        base_dir = os.path.abspath(base_dir) if base_dir else pathlib.Path(".")
        for s in self.all_symbols():
            path = s.get(PATH, None)
            if path:
                str_path = str(path)
                abs_win_path = self.windows_path_pattern.match(str_path)
                if base_dir in path.parents:
                    path = path.relative_to(base_dir)
                # Remove root from path
                elif str_path.startswith("/"):
                    str_path = str_path[1:]
                    path = pathlib.Path(str_path)
                elif abs_win_path:
                    # prefix drive letter 
                    # in the rare case where there are two
                    # files with same path and different drive letter
                    drive_letter = abs_win_path.group(1)
                    str_path = abs_win_path.group(1)+"_"+abs_win_path.group(4)
                    path = pathlib.Path(str_path)
                s[PATH] = path

    def unmangle_cpp_names(self):
        symbol_names = list(symbol[NAME] for symbol in self.all_symbols())

        unmangled_names = self.gcc_tools.get_unmangled_names(symbol_names)

        for s in self.all_symbols():
            s[DISPLAY_NAME] = unmangled_names[s[NAME]]

    def parse_elf(self, elf_file):

        print("parsing ELF at %s" % elf_file)

        self.parse_assembly_text("".join(self.gcc_tools.get_assembly_lines(elf_file)))
        for l in self.gcc_tools.get_size_lines(elf_file):
            self.parse_size_line(l)

        self.elf_mtime = os.path.getmtime(elf_file)

    def parse_su_dir(self, su_dir):

        def gen_find(filepat,top):
            for path, dirlist, filelist in os.walk(top):
                for name in fnmatch.filter(filelist,filepat):
                    yield os.path.join(path,name)

        def gen_open(filenames):
            for name in filenames:
                yield open(name)

        def gen_cat(sources):
            for s in sources:
                for item in s:
                    yield item

        def get_stack_usage_lines(su_dir):
            names = gen_find("*.su", su_dir)
            files = gen_open(names)
            lines = gen_cat(files)
            return lines

        if su_dir:
            print("parsing stack usages starting at %s" % su_dir)
            for l in get_stack_usage_lines(su_dir):
                self.parse_stack_usage_line(l)

    def sorted_by_size(self, symbols):
        return sorted(symbols, key=lambda k: k.get("size", 0), reverse=True)

    def all_symbols(self):
        return self.sorted_by_size(self.symbols.values())

    def all_functions(self):
        return list([f for f in self.all_symbols() if f.get(TYPE, None) == TYPE_FUNCTION])

    def all_variables(self):
        return list([f for f in self.all_symbols() if f.get(TYPE, None) == TYPE_VARIABLE])

    def enhance_assembly(self):
        for key, symbol in self.symbols.items():
            if ASM in symbol:
                symbol[ASM] = list([self.enhanced_assembly_line(l) for l in symbol[ASM]])

    def add_function_call(self, caller, callee):
        if caller != callee:
            if not callee in caller[CALLEES]:
                caller[CALLEES].append(callee)
            if not caller in callee[CALLERS]:
                callee[CALLERS].append(caller)
                caller_file = caller.get("file", None)
                callee_file = callee.get("file", None)
                if callee_file and caller_file and callee_file != caller_file:
                    callee["called_from_other_file"] = True

    #  934:	f7ff bba8 	b.w	88 <jump_to_pbl_function>
    # 8e4:	f000 f824 	bl	930 <app_log>
    #
    # but not:
    # 805bbac:	2471 0805 b64b 0804 b3c9 0804 b459 0804     q$..K.......Y...
    enhance_call_tree_pattern = re.compile(r"^\s*[\da-f]+:\s+[\d\sa-f]{9}\s+BL?(EQ|NE|CS|HS|CC|LO|MI|PL|VS|VC|HI|LS|GE|LT|GT|LE|AL)?(\.W|\.N)?\s+([\d\sa-f]+)", re.IGNORECASE)

    def enhance_call_tree_from_assembly_line(self, function, line):
        if "<" not in line:
            return False

        match = self.enhance_call_tree_pattern.match(line)

        if match:
            callee = self.symbol_by_addr(match.group(3))
            if callee:
                self.add_function_call(function, callee)
                return True

        return False

    def enhance_call_tree(self):
        for f in self.all_functions():
            for k in [CALLERS, CALLEES]:
                f[k] = f.get(k, [])

        for f in self.all_functions():
            if ASM in f:
                [self.enhance_call_tree_from_assembly_line(f, l) for l in f[ASM]]

    def enhance(self, src_root):
        self.normalize_files_paths(src_root)
        print("enhancing function sizes")
        self.enhance_function_size_from_assembly()
        print("deriving folders")
        self.derive_folders()
        print("enhancing file elements")
        self.enhance_file_elements()
        print("enhancing assembly")
        self.enhance_assembly()
        print("enhancing call tree")
        self.enhance_call_tree()
        print("enhancing siblings")
        self.enhance_sibling_symbols()
        self.enhance_symbol_flags()
        print("unmangling c++ symbols")
        self.unmangle_cpp_names()

    #   98: a8a8a8a8  bl 98
    enhanced_assembly_line_pattern = re.compile(r"^\s*[\da-f]+:\s+[\d\sa-f]{9}\s+bl\s+([\d\sa-f]+)\s*$")

    def enhanced_assembly_line(self, line):
        match = self.enhanced_assembly_line_pattern.match(line)
        if match:
            symbol = self.symbol_by_addr(match.group(1))
            if symbol:
                return line+ " <%s>" % (symbol["name"])
        return line

    # 88a:	ebad 0d03 	sub.w	sp, sp, r3
    count_assembly_code_bytes_re = re.compile(r"^\s*[\da-f]+:\s+([\d\sa-f]{9})")

    def count_assembly_code_bytes(self, line):
        match = self.count_assembly_code_bytes_re.match(line)
        if match:
            return len(match.group(1).replace(" ", "")) // 2
        return 0

    def enhance_function_size_from_assembly(self):
        for f in self.all_symbols():
            if ASM in f:
                f[SIZE] = sum([self.count_assembly_code_bytes(l) for l in f[ASM]])

    def enhance_sibling_symbols(self):
        for f in self.all_functions():
            if SIZE in f:
                addr = int(f.get(ADDRESS), 16) + f.get(SIZE)
                next_symbol = self.symbol_by_addr(hex(addr))
                if next_symbol and next_symbol.get(TYPE, None) == TYPE_FUNCTION:
                    f[NEXT_FUNCTION] = next_symbol

        for f in self.all_functions():
            n = f.get(NEXT_FUNCTION, None)
            if n:
                n[PREV_FUNCTION] = f

    def derive_folders(self):
        for s in self.all_symbols():
            unknown_path = pathlib.Path("<unknown>/<unknown>")
            p = s.get(PATH, unknown_path)
            if p != unknown_path:
                posix_root_path = str(p).startswith("\\")
                windows_os = os.name == "nt"
                # Detects if parsing posix paths in elf in a windows machine
                win_parsing_posix = windows_os and posix_root_path
                if not win_parsing_posix:
                    resolved_path = p.resolve(strict=False)
                else:
                    resolved_path = p
                if windows_os and PYTHON_VER["major"]==3 and PYTHON_VER["minor"]<10:
                    pathlib_prepends_cwd = False
                else:
                    pathlib_prepends_cwd = True
                    
                if (not p.is_absolute() 
                    and not win_parsing_posix
                    and pathlib_prepends_cwd): 
                    # pathlib prepends cwd if it couldnt 
                    # resolve locally the file
                    cwd = pathlib.Path().absolute()
                    p = resolved_path.relative_to(cwd)
                else:
                    p = resolved_path
            s[PATH] = p
            s[BASE_FILE] = p.name
            s[FILE] = self.file_for_path(p)
            s[FILE][SYMBOLS].append(s)

    def file_element_for_path(self, path, type, default_values):
        if not path:
            return None

        result = self.file_elements.get(path, None)
        if not result:
            parent_len = len(path.parents)
            if parent_len > 0:
                parent_dir = path.parents[0]
            else:
                parent_dir = path
            parent_available = parent_len > 0 and parent_dir != pathlib.Path(".")
            parent_folder = self.folder_for_path(parent_dir) if parent_available else None 
            result = {
                TYPE: type,
                PATH: path,
                FOLDER: parent_folder,
                NAME: os.path.basename(path),
            }
            for k, v in default_values.items():
                result[k] = v
            self.file_elements[path] = result

        return result if result[TYPE] == type else None

    def file_for_path(self, path):
        return self.file_element_for_path(path, TYPE_FILE, {SYMBOLS:[]})

    def folder_for_path(self, path):
        return self.file_element_for_path(path, TYPE_FOLDER, {FILES:[], SUB_FOLDERS:[], COLLAPSED_SUB_FOLDERS:[]})

    def file_items_ancestors(self, item):
        while item.get(FOLDER):
            item = item[FOLDER]
            yield item

    def enhance_file_elements(self):
        for f in self.all_files():
            parent = f.get(FOLDER, None)
            if parent:
                parent[FILES].append(f)

            f[SYMBOLS] = sorted(f[SYMBOLS], key=lambda s: s[NAME])
            f[FUNCTIONS] = list([s for s in f[SYMBOLS] if s.get(TYPE, None) == TYPE_FUNCTION])
            f[VARIABLES] = list([s for s in f[SYMBOLS] if s.get(TYPE, None) == TYPE_VARIABLE])

        for f in self.all_folders():
            parent = f.get(FOLDER, None)
            if parent:
                parent[SUB_FOLDERS].append(f)
            ancestors = list(self.file_items_ancestors(f))
            f[ANCESTORS] = ancestors
            if len(ancestors) > 0:
                f[ROOT] = ancestors[-1]

            collapsed_name = f[NAME]
            for a in ancestors:
                if len(f[FILES]) > 0:
                    a[COLLAPSED_SUB_FOLDERS].append(f)
                if len(a[FILES]) > 0:
                    break
                collapsed_name = os.path.join(a[NAME], collapsed_name)
            f[COLLAPSED_NAME] = collapsed_name

        for f in self.all_folders():
            for k in [FILES, SUB_FOLDERS]:
                f[k] = sorted(f[k], key=lambda s: s[NAME])
            f[COLLAPSED_SUB_FOLDERS] = sorted(f[COLLAPSED_SUB_FOLDERS], key=lambda s: s[COLLAPSED_NAME])

    def all_files(self):
        return [f for f in self.file_elements.values() if f[TYPE] == TYPE_FILE]

    def all_folders(self):
        return [f for f in self.file_elements.values() if f[TYPE] == TYPE_FOLDER]

    def root_folders(self):
        return [f for f in self.all_folders() if not f[FOLDER]]

    def collapsed_root_folders(self):
        result = []

        def non_empty_leafs(f):
            if len(f[FILES]) > 0:
                result.append(f)
            else:
                for s in f[SUB_FOLDERS]:
                    non_empty_leafs(s)

        for f in self.root_folders():
            non_empty_leafs(f)

        return result

    def enhance_symbol_flags(self):
        is_float_function_pattern = re.compile(r"^__aeabi_(f.*|.*2f)|__addsf3$")
        def is_float_function_name(n):
            return is_float_function_pattern.match(n)

        float_functions = [f for f in self.all_functions() if is_float_function_name(f[NAME])]
        for f in self.all_functions():
            callees = f[CALLEES]
            f["calls_float_function"] = any([ff in callees for ff in float_functions])

        for file in self.all_files():
            file["calls_float_function"] = any([f["calls_float_function"] for f in file[FUNCTIONS]])


        def folder_calls_float_function(folder):
            result = any([f["calls_float_function"] for f in folder[FILES]])
            for sub_folder in folder[SUB_FOLDERS]:
                if folder_calls_float_function(sub_folder):
                    result = True
            folder["calls_float_function"] = result
            return result

        for folder in self.root_folders():
            folder_calls_float_function(folder)

    def build_symbol_name_index(self):
        if not self.symbols_by_name or not self.symbols_by_qualified_name:
            self.symbols_by_name = {}
            self.symbols_by_qualified_name = {}

            for s in self.symbols.values():
                name = s[NAME]
                if name:
                    self.symbols_by_name[name] = s

                qualified_name = self.qualified_symbol_name(s)
                if qualified_name:
                    self.symbols_by_qualified_name[qualified_name] = s



import re
from puncover import collector

class BacktraceHelper():

    def __init__(self, collector):
        self.collector = collector

    derive_functions_symbols_pattern = re.compile(r"\b(\w+)\b")

    def derive_function_symbols(self, text):
        result = []
        for f in self.derive_functions_symbols_pattern.finditer(text):
            s = self.collector.symbol(f.group(1), False)
            if s and s[collector.TYPE] == collector.TYPE_FUNCTION:
                result.append(s)
        return result

    def transform_known_symbols(self, text, transformer):
        def f(match):
            symbol_name = match.group(1)
            symbol = self.collector.symbol(symbol_name, False)
            return transformer(symbol) if symbol else symbol_name

        return self.derive_functions_symbols_pattern.sub(f, text)


    def deepest_call_tree(self, f, list_attribute, cache_attribute, visited = None):
        # TODO: find strongly connected components and count cycles correctly
        if cache_attribute in f:
            return f[cache_attribute]

        visited = [f] + (visited if visited else [])
        result = (0, [])

        for c in f[list_attribute]:
            if c not in visited:
                candidate = self.deepest_call_tree(c, list_attribute, cache_attribute, visited)
                if candidate[0] > result[0]:
                    result = candidate


        result = (result[0] + f.get(collector.STACK_SIZE, 0), [f] + result[1])
        f[cache_attribute] = result
        return result


    def deepest_callee_tree(self, f):
        return self.deepest_call_tree(f, collector.CALLEES, collector.DEEPEST_CALLEE_TREE)

    def deepest_caller_tree(self, f):
        return self.deepest_call_tree(f, collector.CALLERS, collector.DEEPEST_CALLER_TREE)



import abc
import os
from os.path import dirname
from puncover.backtrace_helper import BacktraceHelper
import pathlib

class Builder:

    def __init__(self, collector, src_root):
        self.files = {}
        self.collector = collector
        self.backtrace_helper = BacktraceHelper(collector)
        self.src_root = pathlib.Path(src_root)

    def store_file_time(self, path, store_empty=False):
        self.files[path] = 0 if store_empty else os.path.getmtime(path)

    def build(self):
        for f in self.files.keys():
            self.store_file_time(f)
        self.collector.reset()
        self.collector.parse_elf(self.get_elf_path())
        self.collector.enhance(self.src_root)
        self.collector.parse_su_dir(self.get_su_dir())
        self.build_call_trees()

    def needs_build(self):
        return any([os.path.getmtime(f) > t for f,t in self.files.items()])

    def build_if_needed(self):
        if self.needs_build():
            self.build()

    @abc.abstractmethod
    def get_elf_path(self):
        pass

    @abc.abstractmethod
    def get_su_dir(self):
        pass

    def build_call_trees(self):
        for f in self.collector.all_functions():
            self.backtrace_helper.deepest_callee_tree(f)
            self.backtrace_helper.deepest_caller_tree(f)


class ElfBuilder(Builder):

    def __init__(self, collector, src_root, elf_file, su_dir):
        Builder.__init__(self, collector, src_root if src_root else dirname(dirname(elf_file)))
        self.store_file_time(elf_file, store_empty=True)
        self.elf_file = pathlib.Path(elf_file)
        self.su_dir = su_dir

    def get_elf_path(self):
        return self.elf_file

    def get_su_dir(self):
        return self.su_dir



import os
import subprocess

import itertools


class GCCTools:
    def __init__(self, gcc_base_filename):
        # if base filename is a directory, make sure we have the trailing slash
        if os.path.isdir(gcc_base_filename):
            gcc_base_filename = os.path.join(gcc_base_filename, '')

        self.gcc_base_filename = gcc_base_filename

    def gcc_tool_path(self, name):
        path = self.gcc_base_filename + name
        if os.name == 'nt':
            path+=".exe"
        if not os.path.isfile(path):
            raise Exception("Could not find %s" % path)

        return path

    def gcc_tool_lines(self, name, args, cwd=None):
        proc = subprocess.Popen([self.gcc_tool_path(name)] + args, stdout=subprocess.PIPE, cwd=cwd)
        return [l.decode() for l in proc.stdout.readlines()]

    def get_assembly_lines(self, elf_file):
        return self.gcc_tool_lines('objdump', ['-dslw',  elf_file.name], elf_file.parents[0])

    def get_size_lines(self, elf_file):
        # http://linux.die.net/man/1/nm
        return self.gcc_tool_lines('nm', ['-Sl', elf_file.name], elf_file.parents[0])

    # See https://blog.flameeyes.eu/2010/06/c-name-demangling/ for context
    #
    # This solution courtesy of:
    # https://stackoverflow.com/questions/6526500/c-name-mangling-library-for-python/6526814
    def get_unmangled_names(self, symbol_names, chunk_size=1000):
        # for very long lists we can exceed the maximum length of the command line
        # so we split the names in chunks
        def chunks(l):
            for i in range(0, len(l), chunk_size):
                yield l[i:i + chunk_size]

        lines_list =  [self.gcc_tool_lines('c++filt', c) for c in chunks(symbol_names)]
        lines = list(itertools.chain.from_iterable(lines_list))
        demangled = list(s.rstrip() for s in lines)

        return dict(zip(symbol_names, demangled))



#!/usr/bin/env python

import argparse
import os
import webbrowser
from distutils.spawn import find_executable
from os.path import dirname
from threading import Timer

from flask import Flask

from puncover import renderers
from puncover.builders import ElfBuilder
from puncover.collector import Collector
from puncover.gcc_tools import GCCTools
from puncover.middleware import BuilderMiddleware
from puncover.version import __version__

# Default listening port. Fallback to 8000 if the default port is already in use.
DEFAULT_PORT = 5000
DEFAULT_PORT_FALLBACK = 8000


def is_port_in_use(port: int) -> bool:
    import socket
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


def get_default_port():
    return DEFAULT_PORT if not is_port_in_use(DEFAULT_PORT) else DEFAULT_PORT_FALLBACK


def create_builder(gcc_base_filename, elf_file=None, su_dir=None, src_root=None):
    c = Collector(GCCTools(gcc_base_filename))
    if elf_file:
        return ElfBuilder(c, src_root, elf_file, su_dir)
    else:
        raise Exception("Unable to configure builder for collector")


app = Flask(__name__)


def find_arm_tools_location():
    obj_dump = find_executable("arm-none-eabi-objdump")
    return dirname(dirname(obj_dump)) if obj_dump else None


def open_browser(host, port):
    webbrowser.open("http://{}:{}/".format(host, port))


def main():
    gcc_tools_base = os.path.join(find_arm_tools_location(), 'bin/arm-none-eabi-')

    parser = argparse.ArgumentParser(
        description="Analyses C/C++ build output for code size, static variables, and stack usage.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument('--gcc-tools-base', '--gcc_tools_base', default=gcc_tools_base,
                        help='filename prefix for your gcc tools, e.g. ~/arm-cs-tools/bin/arm-none-eabi-')
    parser.add_argument('--elf_file', '--elf-file', required=True,
                        help='location of an ELF file')
    parser.add_argument('--src_root', '--src-root',
                        help='location of your sources')
    parser.add_argument('--build_dir', '--build-dir',
                        help='location of your build output')
    parser.add_argument('--debug', action='store_true',
                        help='enable Flask debugger')
    parser.add_argument('--port', dest='port', default=get_default_port(), type=int,
                        help='port the HTTP server runs on')
    parser.add_argument('--host', default='127.0.0.1',
                        help='host IP the HTTP server runs on')
    parser.add_argument('--no-open-browser', action='store_true',
                        help="don't automatically open a browser window")
    args = parser.parse_args()

    builder = create_builder(args.gcc_tools_base, elf_file=args.elf_file,
                             src_root=args.src_root, su_dir=args.build_dir)
    builder.build_if_needed()
    renderers.register_jinja_filters(app.jinja_env)
    renderers.register_urls(app, builder.collector)
    app.wsgi_app = BuilderMiddleware(app.wsgi_app, builder)

    if args.debug:
        app.debug = True

    if is_port_in_use(args.port):
        print("Port {} is already in use, please choose a different port.".format(args.port))
        exit(1)

    # Open a browser window, only if this is the first instance of the server
    # from https://stackoverflow.com/a/63216793
    if not args.no_open_browser and not os.environ.get("WERKZEUG_RUN_MAIN"):
        # wait one second before starting, so the flask server is ready and we
        # don't see a 404 for a moment first
        Timer(1, open_browser, kwargs={"host":args.host, "port":args.port}).start()

    app.run(host=args.host, port=args.port)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import os
from pathlib import Path

from setuptools import setup, find_packages, Command

ROOTDIR = Path(__file__).parent

__version__ = None  # Overwritten by executing version.py.
with open(ROOTDIR / "puncover/version.py") as f:
    exec(f.read())


with open(ROOTDIR / "requirements-test.txt") as f:
    tests_require = list(filter(lambda x: not x.strip().startswith('-r'), f.readlines()))

with open(ROOTDIR / "requirements.txt") as f:
    requires = f.readlines()


class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""

    # http://stackoverflow.com/a/3780822/196350
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system("rm -vrf ./build ./dist ./*.pyc ./*.tgz ./*.egg-info")


setup(
    name="puncover",
    version=__version__,
    description="Analyses C/C++ build output for code size, static variables, and stack usage.",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    url="https://github.com/hbehrens/puncover",
    download_url="https://github.com/hbehrens/puncover/tarball/%s" % __version__,
    author="Heiko Behrens",
    maintainer="Victor Chavez",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    packages=find_packages(exclude=["tests", "tests.*"]),
    include_package_data=True,
    zip_safe=False,
    entry_points={"console_scripts": ["puncover = puncover.puncover:main"]},
    install_requires=requires,
    tests_require=tests_require,
    cmdclass={
        "clean": CleanCommand,
    },
    # TODO: https://github.com/HBehrens/puncover/issues/36
    #  Fix Python 3.5
    python_requires=">=3.6",
)
